/*
    This file is part of magpar.

    Copyright (C) 2002-2008 Werner Scholz

    www:   http://magnet.atp.tuwien.ac.at/scholz/magpar/
    email: magpar(at)magnet.atp.tuwien.ac.at

    magpar is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    magpar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with magpar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

static char const Id[] = "$Id: htot.c 2342 2008-12-11 04:34:12Z scholz $\n\n";
static char const Td[] = "$Today: " __FILE__ "  " __DATE__ "  "  __TIME__  " $\n\n";


#include "griddata.h"
#include "field.h"

#ifdef ADDONS
#include "addons/addons.h"
#endif

#ifdef SPINTORQ
#include "spintorq/spintorq.h"
#endif

static int doinit=1;
static Vec VHtotsum=PETSC_NULL;
static Vec VMom; /**< magnetic moment: VMom=M*Vms3 , energy: E=H*VMom */

int Htot_Init(GridData *gdata)
{
  MagparFunctionLogBegin;

  PetscPrintf(PETSC_COMM_WORLD,"Htot on\n");

  ierr = VecDuplicate(gdata->M,&gdata->VHtot);CHKERRQ(ierr);
  ierr = VecDuplicate(gdata->M,&VMom);CHKERRQ(ierr);
  ierr = VecDuplicate(gdata->M,&VHtotsum);CHKERRQ(ierr);

  MagparFunctionLogReturn(0);
}


int Htot(GridData *gdata)
{
  MagparFunctionInfoBegin;

  /* initialize if necessary */
  if (doinit) {
    ierr = Htot_Init(gdata);CHKERRQ(ierr);
    doinit=0;
  }

#if PETSC_VERSION >= 230
  ierr = VecZeroEntries(VHtotsum);CHKERRQ(ierr);
#else
  const PetscReal  c_zero=0.0;
  ierr = VecSet(&c_zero,VHtotsum);CHKERRQ(ierr);
#endif

  ierr = Hcubic(gdata,VHtotsum);CHKERRQ(ierr);
  ierr = Hdemag(gdata,VHtotsum);CHKERRQ(ierr);
  ierr = Helastic(gdata,VHtotsum);CHKERRQ(ierr);
  ierr = Hexchani(gdata,VHtotsum);CHKERRQ(ierr);
  ierr = Hexternal(gdata,VHtotsum);CHKERRQ(ierr);
#ifdef ADDONS
  ierr = HIlayerExch(gdata,VHtotsum);CHKERRQ(ierr);
  ierr = Hnonlin(gdata,VHtotsum);CHKERRQ(ierr);
#endif
#ifdef SPINTORQ
  /* initialize vectors to NULL */
  gdata->VHspintorque=PETSC_NULL;
  gdata->VHspinforce=PETSC_NULL;

  ierr = Hspintorq(gdata);CHKERRQ(ierr);
  ierr = Hspinforce(gdata);CHKERRQ(ierr);

  if (gdata->VHspintorque!=PETSC_NULL && gdata->VHspinforce!=PETSC_NULL) {
    SETERRQ(PETSC_ERR_ARG_INCOMP,"Spin torque and spin force are switched on at the same time!\n");
  }
#endif

  ierr = VecCopy(VHtotsum,gdata->VHtot);CHKERRQ(ierr);

  MagparFunctionProfReturn(0);
}


int Htot_Gradient(GridData *gdata)
{
  Vec VHtemp;

  MagparFunctionInfoBegin;

  /* abusing VMom */
  VHtemp=VMom;

  /* initialize if necessary */
  if (doinit) {
    ierr = Htot_Init(gdata);CHKERRQ(ierr);
    doinit=0;
  }

  ierr = VecZeroEntries(VHtotsum);CHKERRQ(ierr);

#if 1
  ierr = Htot(gdata);CHKERRQ(ierr);
#else
  /* calculate field/gradient contributions and apply fudge factor */
  VecZeroEntries(VHtemp);Hcubic(gdata,VHtemp);     VecAXPY(VHtotsum,2.0,VHtemp); /* ok */
  VecZeroEntries(VHtemp);Hdemag(gdata,VHtemp);     VecAXPY(VHtotsum,1.0,VHtemp); /* ok */
  VecZeroEntries(VHtemp);Helastic(gdata,VHtemp);   VecAXPY(VHtotsum,4.0,VHtemp); /* ok */
  VecZeroEntries(VHtemp);Hexchani(gdata,VHtemp);   VecAXPY(VHtotsum,2.0,VHtemp); /* ok */
  VecZeroEntries(VHtemp);Hexternal(gdata,VHtemp);  VecAXPY(VHtotsum,2.0,VHtemp); /* ok */
#ifdef ADDONS
  VecZeroEntries(VHtemp);HIlayerExch(gdata,VHtemp);VecAXPY(VHtotsum,2.0,VHtemp); /* TODO: to be verified */
  VecZeroEntries(VHtemp);Hnonlin(gdata,VHtemp);    VecAXPY(VHtotsum,1.0,VHtemp); /* ok */
#endif

  ierr = VecCopy(VHtotsum,gdata->VHtot);CHKERRQ(ierr);
#endif

  /* scale gradient to get agreement between hand-coded and finite difference gradient */
  ierr = VecScale(gdata->VHtot,3.0/2.0);CHKERRQ(ierr);

  MagparFunctionProfReturn(0);
}


int Htot_Energy(GridData *gdata)
{
  MagparFunctionInfoBegin;

  /* calculate "magnetic moments" VMom on vertices */
#if PETSC_VERSION >= 230
  ierr = VecPointwiseMult(VMom,gdata->VMs3,gdata->M);CHKERRQ(ierr);
#else
  ierr = VecPointwiseMult(gdata->VMs3,gdata->M,VMom);CHKERRQ(ierr);
#endif

  PetscReal e,etot;
  etot=0;

  ierr = Hcubic_Energy(gdata,VMom,&e);     CHKERRQ(ierr);etot += e;gdata->Eexchani=e/gdata->totvol; /* part of anisotropy energy */
  ierr = Hdemag_Energy(gdata,VMom,&e);     CHKERRQ(ierr);etot += e;gdata->Edem=e/gdata->totvol;
  ierr = Helastic_Energy(gdata,VMom,&e);   CHKERRQ(ierr);etot += e;
  ierr = Hexchani_Energy(gdata,VMom,&e);   CHKERRQ(ierr);etot += e;gdata->Eexchani += e/gdata->totvol; /* add to anisotropy energy */
#ifdef EXCH
  ierr = Hexchonly_Energy(gdata,VMom,&e);  CHKERRQ(ierr);etot += 0;gdata->Eexchonly=e/gdata->totvol; /* don't add to etot!!! */
#endif
  ierr = Hexternal_Energy(gdata,VMom,&e);  CHKERRQ(ierr);etot += e;gdata->Eext=e/gdata->totvol;
#ifdef ADDONS
  ierr = HIlayerExch_Energy(gdata,VMom,&e);CHKERRQ(ierr);etot += e;gdata->Eexchani += e/gdata->totvol; /* add to anisotropy+exchange energy */
  ierr = Hnonlin_Energy(gdata,VMom,&e);   CHKERRQ(ierr);etot += e;
#ifdef EXCH
  gdata->Eexchonly += e/gdata->totvol; /* add to exchange only energy */
#endif
#endif
#ifdef SPINTORQ
/* TODO: implement
  ierr = Hspintorq_Energy(gdata,VMom,&e);  CHKERRQ(ierr);etot += e;
  ierr = Hspinforce_Energy(gdata),VMom,&e; CHKERRQ(ierr);etot += e;
*/
#endif

  gdata->Etot=etot/gdata->totvol;

  MagparFunctionProfReturn(0);
}

